-- the app size
local myappwidth, myappheight = application:getContentWidth(), application:getContentHeight()
local matrix = Matrix.new()

-- react physics 3D
r3d = require "reactphysics3d"
local world = r3d.World.new(0, -9.8, 0)
world.staticbodies = {} -- staticbodies list
world.cubes = {} -- cubes list
world.objs = {} -- objs list
world.missiles = {} -- missiles list
-- a fullscreen 3D viewport
local view = D3.View.new(myappwidth, myappheight, 45, 0.1, 2048 * 3) -- function D3View:init(width, height, fov, near, far)

-- keyboard control
local isup, isdown, isleft, isright, isshoot, isjump = false, false, false, false, false, false

-- a sky
local sky = Pixel.new(0xFFFFFF, 1, myappwidth * 2, myappheight * 2)
sky:setAnchorPoint(0.5, 0.5)
sky:setPosition(myappwidth / 2, myappheight / 2)
sky:setColor(0x00FFFF, 1, 0x00000FF, 1, 90) -- 0x00040FF -- blue gradient background
stage:addChild(sky)

-- some ground
local groundsize = 64 * 64
local ground01size = 4 * 64
local ground = PLANE_3D(world, { sizex=groundsize, sizez=groundsize, texpath="3d_objs/texs/groundx01.png", texw=96*64, texh=96*64 })
local ground01 = PLANE_3D(world, { posy=6, sizex=ground01size, sizez=ground01size, texpath="3d_objs/texs/grass.png", texw=32*64, texh=32*64 })
-- some walls
local wallheight = 64
PLANE_3D( world, { posy=wallheight, posz=-groundsize, sizex=groundsize, sizez=wallheight, rotx=90, texpath="3d_objs/texs/box.png", texw=16*80, texh=8*80 } )
PLANE_3D( world, { posy=wallheight, posz=groundsize, sizex=groundsize, sizez=wallheight, rotx=-90, texpath="3d_objs/texs/box.png", texw=16*80, texh=8*80 } )
PLANE_3D( world, { posx=groundsize, posy=wallheight, sizex=groundsize, sizez=wallheight, rotx=90, rotz=90, texpath="3d_objs/texs/box.png", texw=16*80, texh=8*80 } )
PLANE_3D( world, { posx=-groundsize, posy=wallheight, sizex=groundsize, sizez=wallheight, rotx=90, rotz=-90, texpath="3d_objs/texs/box.png", texw=16*80, texh=8*80 } )
-- some ramps
PLANE_3D( world, { posy=6, posz=-ground01size + 16, sizex=32, sizez=128, rotx=15, texpath="3d_objs/texs/box.png", texw=16*80, texh=12*80 } )
PLANE_3D( world, { posy=4, posz=ground01size - 16, sizex=48, sizez=192, rotx=-12.5, roty=0, texpath="3d_objs/texs/box.png", texw=16*80, texh=12*80 } )
-- some cubes
local cube01, cube01body = cube_factory( world, { sizex = 8, posx = 1*64, posy = 8 * 8, posz = -128, mass = 8 * 80 } )
local cube02, cube02body = cube_factory( world, { sizex = 4, posx = -2*64, posy = 4 * 4, posz = -96, mass = 4 * 40 } )
-- the player
local shipobj, shipbody = PLAYER(world)
-- some big ships
local mereobj, merebody = obj_factory(world, "3d_objs/cruiser", "vaisseau_mere5.obj", { posx=0, posy=196/2, posz=128, rotx=0, roty=0, rotz=0, mass=16*1024, gravity=false } )
local mere02obj, mere02body = obj_factory(world, "3d_objs/cruiser", "vaisseau_mere5.obj", { posx=233 , posy=100, posz=512, rotx=0, roty=0, rotz=0, mass=16*1024, gravity=false } )
local mere03obj, mere03body = obj_factory(world, "3d_objs/cruiser", "vaisseau_mere5.obj", { posx=-233, posy=100, posz=512, rotx=0, roty=0, rotz=0, mass=16*1024, gravity=false } )

-- event listener
function collisionsEventListener()
	for k, v in pairs(world.missiles) do
		for k1, v1 in pairs(world.cubes) do
			if world:testAABBOverlap(v.xbody, v1.xbody) then
				v.isdirty = true
			end
		end
	end
end
world:setEventListener(collisionsEventListener)

-- the scene
local scene = view:getScene() -- a 3D View sprite
--view:lookAt(0, 10, -20, 0, 5, 0) -- initial camera lookat 0, 10, -20, 0, 5, 0
stage:addChild(view)
-- add the objects
for p = 1, #world.staticbodies do scene:addChild(world.staticbodies[p]) end
for c = 1, #world.cubes do scene:addChild(world.cubes[c].xobj) end
for o = 1, #world.objs do scene:addChild(world.objs[o].xobj) end
scene:addChild(shipobj)
-- the light
Lighting.setLight(1 * 64, 8 * 64, 0, 0.2) -- 15, 30, 0, 0.3
Lighting.setLightTarget(0, 0, -64, 64 * 64, 1024) -- 0, 0, 0, 40, 120

-- the game loop
matrix = shipbody:getTransform() -- position and rotation matrix
local ax, ay, az = matrix:transformPoint(0, 0, 0) -- to calculate angles
local bx, by, bz = matrix:transformPoint(0, 0, -1) -- -1 because by default the ship is facing the camera
local dx, dy, dz = bx - ax, by - ay, bz - az -- the angles
local shipweight = 3 * shipobj.frontmass + 1 * shipobj.rearmass -- because ship has 3 front objects and 1 rear object
local camx, camy, camz -- a camera
local missilelimit = 256 -- the missiles max limit before being removed from the world and the scene
stage:addEventListener(Event.ENTER_FRAME, function(e)
	-- lighting
	Lighting.computeShadows(scene)
	-- destroy
	for k, v in pairs(world.missiles) do
		if math.abs(shipobj:getX() - v.xobj:getX()) >= missilelimit or
				math.abs(shipobj:getY() - v.xobj:getY()) >= missilelimit or
				math.abs(shipobj:getZ() - v.xobj:getZ()) >= missilelimit or
				v.isdirty then
			world.missiles[k] = nil
			world:destroyBody(v.xbody)
			scene:removeChild(v.xobj)
		end
	end
	-- action!
	world:step(e.deltaTime)
	-- the boxes
	for _, v in pairs(world.cubes) do
		local m1 = v.xbody:getTransform() -- not the matrix
		v.xobj:setMatrix(m1)
	end
	-- the meres
	for _, v in pairs(world.objs) do
		v.xbody:applyForce(0, 0, - 1.5 * 1024)
		local m4 = v.xbody:getTransform() -- not the matrix
		v.xobj:setMatrix(m4)
	end
	-- the ship
	matrix = shipbody:getTransform()
	ax, ay, az = matrix:transformPoint(0, 0, 0)
	bx, by, bz = matrix:transformPoint(0, 0, -1) -- -1 because by default the ship is facing the camera
	dx, dy, dz = bx - ax, by - ay, bz - az
	if isup then shipbody:applyForce(dx * shipweight * 14, 0, dz * shipweight * 14) end
	if isdown then shipbody:applyForce(-dx * shipweight * 14, 0, -dz * shipweight * 14) end
	if isleft then shipbody:applyTorque(0, -shipweight * 40, 0) end
	if isright then shipbody:applyTorque(0, shipweight * 40, 0) end
	if isshoot then shoot() isshoot = false end
	if isjump then shipbody:applyForce(0, shipweight * 32, 0) end
	shipobj:setMatrix(matrix)
	-- the missiles
	for _, v in pairs(world.missiles) do
		matrix = v.xbody:getTransform()
		v.xbody:applyForce(v.xdx * 8 * 64, v.xdy * 8 * 64, v.xdz * 8 * 64)
--		v.xbody:applyForce(v.xdx * 8 * 64, 0, v.xdz * 8 * 64)
		v.xbody:applyTorque(0, 0, 1)
		-- move the obj with its body
		v.xobj:setPosition(matrix:getPosition())
	end
	-- the camera FPS style
	camx, camy, camz = shipobj:getX(), shipobj:getY(), shipobj:getZ()
	camx += -dx * 24 -- 16
	camy += 4 -- 4
	camz += -dz * 24 -- 16
	view:lookAt(camx, camy, camz, shipobj:getX() + 0, shipobj:getY() + 6, shipobj:getZ() + 0,	0, 1, 0)
	-- lighting
--	LightingV1.setLight(camx, camy, camz, 0.5)
end)

-- shoot
function shoot()
	scene:addChild(obj_factory( world, "3d_objs/ship", "missile.obj",
		{ posx=shipobj:getX() + 0, posy=shipobj:getY() + 2.5, posz=shipobj:getZ() + 0,
		rotx=dx, roty=dy,rotz=dz, ismissile=true, shooter=shipobj, mass=10 } )
	)
end

-- keyboard
function onKeyDown(e)
	if e.keyCode == KeyCode.UP then isup = true end
	if e.keyCode == KeyCode.DOWN then isdown = true end
	if e.keyCode == KeyCode.LEFT then isleft = true end
	if e.keyCode == KeyCode.RIGHT then isright = true end
	if e.keyCode == KeyCode.W then isshoot = true end
	if e.keyCode == KeyCode.X then isjump = true end
end
function onKeyUp(e)
	if e.keyCode == KeyCode.UP then isup = false end
	if e.keyCode == KeyCode.DOWN then isdown = false end
	if e.keyCode == KeyCode.LEFT then isleft = false end
	if e.keyCode == KeyCode.RIGHT then isright = false end
	if e.keyCode == KeyCode.W then isshoot = false end
	if e.keyCode == KeyCode.X then isjump = false end
end
stage:addEventListener(Event.KEY_DOWN, onKeyDown)
stage:addEventListener(Event.KEY_UP, onKeyUp)
